@@ -15,7 +15,7 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
 # pylint: disable=invalid-name, too-few-public-methods, too-many-arguments
-# pylint: disable=missing-docstring
+# pylint: disable=missing-docstring, too-many-lines
 """
 Python ctypes bindings for libusb-1.0.
 
@@ -26,13 +26,15 @@ Some are only available after calling loadLibrary.
 """
 from ctypes import (
     Structure, LittleEndianStructure,
+    Union,
     CFUNCTYPE, POINTER, addressof, sizeof, cast,
-    c_short, c_int, c_uint, c_long,
+    c_short, c_int, c_uint, c_long, c_longlong,
     c_uint8, c_uint16, c_uint32,
     c_void_p, c_char_p, py_object, pointer, c_char,
     c_ssize_t, CDLL
 )
 import ctypes.util
+import errno
 import os.path
 import platform
 import sys
@@ -54,9 +56,10 @@ class Enum:
                 next_value += 1
             forward_dict[name] = value
             if value in reverse_dict:
-                raise ValueError('Multiple names for value %r: %r, %r' % (
-                    value, reverse_dict[value], name
-                ))
+                raise ValueError(
+                    'Multiple names for value '
+                    f'{value!r}: {reverse_dict[value]!r}, {name!r}'
+                )
             reverse_dict[value] = name
             scope_dict[name] = value
         self.forward_dict = forward_dict
@@ -118,6 +121,11 @@ def newDescriptor(field_name_list):
     """
     return newStruct(['bLength', 'bDescriptorType'] + list(field_name_list))
 
+# Stand-in for the function until libusb is loaded - if the current libusb
+# version has this export.
+def libusb_strerror(_unused_errcode):
+    return None
+
 class USBError(Exception):
     value = None
 
@@ -127,11 +135,29 @@ class USBError(Exception):
             self.value = value
 
     def __str__(self):
-        return '%s [%s]' % (libusb_error.get(self.value, 'Unknown error'),
-                            self.value)
+        return f'{libusb_error.get(self.value, "Unknown error")} [{self.value}]'
+
+    def getMessage(self):
+        """
+        Get user-friendly message representing the current error, and based on
+        the current locale.
+
+        Returns str, or None.
+        If libusb has not been loaded yet, returns None.
+        """
+        # pylint: disable=assignment-from-none
+        message = libusb_strerror(self.value)
+        # pylint: enable=assignment-from-none
+        if message is not None:
+            message = message.decode('utf-8')
+        return message
 
 c_uchar = c_uint8
 c_int_p = POINTER(c_int)
+intptr_t = {
+    sizeof(x): x
+    for x in (c_int, c_long, c_longlong)
+}[sizeof(c_void_p)]
 
 LITTLE_ENDIAN = sys.byteorder == 'little'
 
@@ -140,48 +166,70 @@ class timeval(Structure):
                 ('tv_usec', c_long)]
 timeval_p = POINTER(timeval)
 
+if platform.system() == 'Windows':
+    LIBUSB_CALL_FUNCTYPE = ctypes.WINFUNCTYPE
+else:
+    LIBUSB_CALL_FUNCTYPE = CFUNCTYPE
+
 def __getLibrary():
+    my_dir = os.path.dirname(__file__)
     system = platform.system()
+    # If this is a binary wheel, try to use an integrated libusb first.
+    # To use the libusb from the Python installation or the OS, install
+    # from sdist:
+    #   > pip install --no-binary :all: libusb1
     if system == 'Windows':
         dll_loader = ctypes.WinDLL
-        suffix = '.dll'
+        libusb_list = [
+            os.path.join(my_dir, 'libusb-1.0.dll'),
+            'libusb-1.0.dll',
+        ]
+        find_library = None
     else:
-        dll_loader = ctypes.CDLL
-        suffix = '.dylib' if system == 'Darwin' else '.so'
-    filename = 'libusb-1.0' + suffix
-    # If this is a binary wheel, use the integrated libusb unconditionally.
-    # To use the libusb from the Python installation or the OS, install from sdist:
-    #   > pip install --no-binary :all: libusb1
-    if os.path.exists(os.path.join(os.path.dirname(__file__), filename)):
-        filename = os.path.join(os.path.dirname(__file__), filename)
-    try:
-        return dll_loader(filename, use_errno=True, use_last_error=True)
-    except OSError:
-        libusb_path = None
-        base_name = 'usb-1.0'
-        if 'FreeBSD' in system:
-            # libusb.so.2 on FreeBSD: load('libusb.so') would work fine, but...
-            # libusb.so.2debian on Debian GNU/kFreeBSD: here it wouldn't work.
-            # So use find_library instead.
-            base_name = 'usb'
-        elif system == 'Darwin':
-            for libusb_path in (
-                    # macport standard path
-                    '/opt/local/lib/libusb-1.0.dylib',
-                    # fink standard path
-                    '/sw/lib/libusb-1.0.dylib',
-                    # homebrew standard path for symlink (Apple M1 Silicon)
-                    '/opt/homebrew/opt/libusb/lib/libusb-1.0.dylib',
-                ):
-                if os.path.exists(libusb_path):
-                    break
-            else:
-                libusb_path = None
-        if libusb_path is None:
-            libusb_path = ctypes.util.find_library(base_name)
-            if libusb_path is None:
-                raise
-        return dll_loader(libusb_path, use_errno=True, use_last_error=True)
+        dll_loader = CDLL
+        if system == 'Darwin':
+            libusb_list = [
+                os.path.join(my_dir, 'libusb-1.0.dylib'),
+                'libusb-1.0.dylib',
+                # macport standard path
+                '/opt/local/lib/libusb-1.0.dylib',
+                # fink standard path
+                '/sw/lib/libusb-1.0.dylib',
+                # homebrew standard path for symlink (Apple M1 Silicon)
+                '/opt/homebrew/opt/libusb/lib/libusb-1.0.dylib',
+            ]
+            find_library = None
+        else:
+            # .so.0 should be the optimal suffix
+            # .so is for BBB, especially if libusb-1.0.so was bundled in some
+            # uses of this module.
+            libusb_list = [
+                os.path.join(my_dir, 'libusb-1.0.so.0'),
+                os.path.join(my_dir, 'libusb-1.0.so'),
+                'libusb-1.0.so.0',
+                'libusb-1.0.so',
+            ]
+            find_library = (
+                # libusb.so.2 on FreeBSD: load('libusb.so') would work fine, but...
+                # libusb.so.2debian on Debian GNU/kFreeBSD: here it wouldn't work.
+                'usb'
+                if 'FreeBSD' in system else
+                'usb-1.0'
+            )
+    for filename in libusb_list:
+        try:
+            return dll_loader(filename, use_errno=True, use_last_error=True)
+        except OSError:
+            pass
+    if find_library is not None:
+        filename = ctypes.util.find_library(find_library)
+        if filename is not None:
+            return dll_loader(filename, use_errno=True, use_last_error=True)
+    raise OSError(
+        errno.ENOENT,
+        'cannot find a suitable libusb-1.0',
+        libusb_list,
+    )
 
 __load_lock = Lock()
 __loaded = False
@@ -248,9 +296,30 @@ def __loadLibrary(libusb): # pylint: disable=too-many-locals,too-many-branches,t
     libusb_transfer._fields_ = _libusb_transfer_fields
     # pylint: enable=protected-access
 
-    #int libusb_init(libusb_context **ctx);
-    libusb_init = libusb.libusb_init
-    libusb_init.argtypes = [libusb_context_p_p]
+    #int libusb_init_context(libusb_context **ctx,
+    #        const struct libusb_init_option options[], int num_options);
+    try:
+        libusb_init_context = libusb.libusb_init_context
+    except AttributeError:
+        # Note: libusb_init is deprecated, only retrieve it when
+        # libusb_init_context is absent.
+        #int libusb_init(libusb_context **ctx);
+        libusb_init = libusb.libusb_init
+        libusb_init.argtypes = [libusb_context_p_p]
+        def libusb_init_context(ctx, _unused_options, num_options):
+            if num_options:
+                raise ValueError(
+                    'libusb_init_context is not available in the current '
+                    'libusb version, initialisation options are not available',
+                )
+            return libusb_init(ctx)
+    else:
+        libusb_init_context.argtypes = [
+            libusb_context_p_p,
+            POINTER(libusb_init_option),
+            c_int,
+        ]
+
     #void libusb_exit(libusb_context *ctx);
     libusb_exit = libusb.libusb_exit
     libusb_exit.argtypes = [libusb_context_p]
@@ -259,6 +328,15 @@ def __loadLibrary(libusb): # pylint: disable=too-many-locals,too-many-branches,t
     libusb_set_debug = libusb.libusb_set_debug
     libusb_set_debug.argtypes = [libusb_context_p, c_int]
     libusb_set_debug.restype = None
+    #void libusb_set_log_cb(libusb_context *ctx, libusb_log_cb cb, int mode);
+    try:
+        libusb_set_log_cb = libusb.libusb_set_log_cb
+    except AttributeError:
+        def libusb_set_log_cb(ctx, cb, mode):
+            pass
+    else:
+        libusb_set_log_cb.argtypes = [libusb_context_p, libusb_log_cb_p, c_int]
+        libusb_set_log_cb.restype = None
     #const struct libusb_version * libusb_get_version(void);
     try:
         libusb_get_version = libusb.libusb_get_version
@@ -292,6 +370,25 @@ def __loadLibrary(libusb): # pylint: disable=too-many-locals,too-many-branches,t
     else:
         libusb_error_name.argtypes = [c_int]
         libusb_error_name.restype = c_char_p
+    #int libusb_setlocale(const char *locale);
+    try:
+        libusb_setlocale = libusb.libusb_setlocale
+    except AttributeError:
+        def libusb_setlocale(_unused_locale):
+            pass
+    else:
+        libusb_setlocale.argtypes = [c_char_p]
+        libusb_setlocale.restype = int
+    #const char *libusb_strerror(int errcode);
+    try:
+        # pylint: disable=redefined-outer-name
+        libusb_strerror = libusb.libusb_strerror
+        # pylint: enable=redefined-outer-name
+    except AttributeError:
+        pass
+    else:
+        libusb_strerror.argtypes = [c_int]
+        libusb_strerror.restype = c_char_p
     #ssize_t libusb_get_device_list(libusb_context *ctx,
     #        libusb_device ***list);
     libusb_get_device_list = libusb.libusb_get_device_list
@@ -362,7 +459,7 @@ def __loadLibrary(libusb): # pylint: disable=too-many-locals,too-many-branches,t
         libusb_get_port_numbers.restype = c_int
     # Missing: libusb_get_port_path (deprecated since 1.0.16)
     try:
-        #libusb_device * LIBUSB_CALL libusb_get_parent(libusb_device *dev);
+        #libusb_device *libusb_get_parent(libusb_device *dev);
         libusb_get_parent = libusb.libusb_get_parent
     except AttributeError:
         pass
@@ -401,6 +498,23 @@ def __loadLibrary(libusb): # pylint: disable=too-many-locals,too-many-branches,t
     else:
         libusb_get_max_iso_packet_size.argtypes = [libusb_device_p, c_uchar]
 
+    #int libusb_wrap_sys_device(libusb_context *ctx, intptr_t sys_dev,
+    #        libusb_device_handle **dev_handle);
+    try:
+        libusb_wrap_sys_device = libusb.libusb_wrap_sys_device
+    except AttributeError:
+        # pylint: enable=unused-argument
+        def libusb_wrap_sys_device(_, __, ___):
+            raise NotImplementedError
+        # pylint: disable=unused-argument
+    else:
+        libusb_wrap_sys_device.restype = c_int
+        libusb_wrap_sys_device.argtypes = [
+            libusb_context_p,
+            intptr_t,
+            libusb_device_handle_p_p,
+        ]
+
     #int libusb_open(libusb_device *dev, libusb_device_handle **handle);
     libusb_open = libusb.libusb_open
     libusb_open.argtypes = [libusb_device_p, libusb_device_handle_p_p]
@@ -634,6 +748,8 @@ def __loadLibrary(libusb): # pylint: disable=too-many-locals,too-many-branches,t
         ]
         libusb_hotplug_deregister_callback.restype = None
 
+    # Note: no support possible for libusb_set_option as it uses a variadic
+
     globals().update(locals())
 
 # libusb.h
@@ -935,6 +1051,7 @@ libusb_config_descriptor_p = POINTER(libusb_config_descriptor)
 libusb_config_descriptor_p_p = POINTER(libusb_config_descriptor_p)
 
 class libusb_control_setup(Structure):
+    _pack_ = 1
     _fields_ = [
         ('bmRequestType', c_uint8),
         ('bRequest', c_uint8),
@@ -1007,6 +1124,22 @@ class libusb_version(Structure):
         ('describe', c_char_p),
     ]
 
+#typedef void (*libusb_log_cb)(libusb_context *ctx,
+#        enum libusb_log_level level, const char *str);
+libusb_log_cb_p = LIBUSB_CALL_FUNCTYPE(None, libusb_context_p, c_int, c_char_p)
+
+class libusb_init_option_value(Union):
+    _fields_ = [
+        ('ival', c_int),
+        ('log_cbval', libusb_log_cb_p),
+    ]
+
+class libusb_init_option(Structure):
+    _fields_ = [
+        ('option', c_int),
+        ('value', libusb_init_option_value),
+    ]
+
 libusb_speed = Enum({
     # The OS doesn't report or know the device speed.
     'LIBUSB_SPEED_UNKNOWN': 0,
@@ -1113,7 +1246,7 @@ class libusb_transfer(Structure):
     pass
 libusb_transfer_p = POINTER(libusb_transfer)
 
-libusb_transfer_cb_fn_p = CFUNCTYPE(None, libusb_transfer_p)
+libusb_transfer_cb_fn_p = LIBUSB_CALL_FUNCTYPE(None, libusb_transfer_p)
 
 libusb_capability = Enum({
     # The libusb_has_capability() API is available.
@@ -1134,17 +1267,18 @@ libusb_log_level = Enum({
     'LIBUSB_LOG_LEVEL_DEBUG': 4,
 })
 
-# Note on libusb_strerror, libusb_setlocale and future functions in the
-# same spirit:
-# I do not think end-user-facing messages belong to a technical library.
-# Such features bring a new, non essential set of problems, and is a luxury
-# I do not want to spend time supporting considering limited resources and
-# more important stuff to work on.
-# For backward compatibility, expose libusb_strerror placeholder.
-# pylint: disable=unused-argument
-def libusb_strerror(errcode):
-    return None
-# pylint: enable=unused-argument
+libusb_log = Enum({
+    'LIBUSB_LOG_CB_GLOBAL': 1 << 0,
+    'LIBUSB_LOG_CB_CONTEXT': 1 << 1,
+})
+
+libusb_option = Enum({
+    'LIBUSB_OPTION_LOG_LEVEL': 0,
+    'LIBUSB_OPTION_USE_USBDK': 1,
+    'LIBUSB_OPTION_NO_DEVICE_DISCOVERY': 2,
+    'LIBUSB_OPTION_LOG_CB': 3,
+    'LIBUSB_OPTION_MAX': 4,
+})
 
 # Get the data section of a control transfer. This convenience function is here
 # to remind you that the data does not start until 8 bytes into the actual
@@ -1290,9 +1424,7 @@ def get_extra(descriptor):
             length = extra[0]
             if not 0 < length <= len(extra):
                 raise ValueError(
-                    'Extra descriptor %i is incomplete/invalid' % (
-                        len(result),
-                    ),
+                    f'Extra descriptor {len(result)} is incomplete/invalid',
                 )
             append(extra[:length])
             extra = extra[length:]
@@ -1349,8 +1481,8 @@ class libusb_pollfd(Structure):
 libusb_pollfd_p = POINTER(libusb_pollfd)
 libusb_pollfd_p_p = POINTER(libusb_pollfd_p)
 
-libusb_pollfd_added_cb_p = CFUNCTYPE(None, c_int, c_short, py_object)
-libusb_pollfd_removed_cb_p = CFUNCTYPE(None, c_int, py_object)
+libusb_pollfd_added_cb_p = LIBUSB_CALL_FUNCTYPE(None, c_int, c_short, py_object)
+libusb_pollfd_removed_cb_p = LIBUSB_CALL_FUNCTYPE(None, c_int, py_object)
 
 #typedef int libusb_hotplug_callback_handle;
 libusb_hotplug_callback_handle = c_int
@@ -1364,11 +1496,12 @@ libusb_hotplug_event = Enum({
     'LIBUSB_HOTPLUG_EVENT_DEVICE_LEFT': 0x02,
 })
 
+LIBUSB_HOTPLUG_NO_FLAGS = 0
 LIBUSB_HOTPLUG_MATCH_ANY = -1
 
 #typedef int (*libusb_hotplug_callback_fn)(libusb_context *ctx,
 #        libusb_device *device, libusb_hotplug_event event, void *user_data);
-libusb_hotplug_callback_fn_p = CFUNCTYPE(
+libusb_hotplug_callback_fn_p = LIBUSB_CALL_FUNCTYPE(
     c_int, libusb_context_p, libusb_device_p, c_int, c_void_p)
 
 # /libusb.h
